import sys
from numpy import sin
from scipy.optimize import minimize
import matplotlib.pyplot as plt


#nelder-mead    Downhill simplex
#powell         Modified Powell
#cg             conjugate gradient (Polak-Ribiere method)
#bfgs           BFGS法
#newton-cg      Newton-CG
#trust-ncg      信頼領域 Newton-CG 法
#dogleg         信頼領域 dog-leg 法
#L-BFGS-B’ (see here)
#TNC’ (see here)
#COBYLA’ (see here)
#SLSQP’ (see here)
#trust-constr’(see here)
#dogleg’ (see here)
#trust-exact’ (see here)
#trust-krylov’ (see here)

method = "bfgs"


#==========================================
# Source parameters to be fitted
#==========================================
x0s = [0.0]

# 数値微分する際の変数の微小変位
h = 0.01

maxiter = 100
tol = 1.0e-5


#==========================================
# Graph parameters
#==========================================
fplot  = 1
ngdata = 51
xgmin  = -4.0
xgmax  =  4.0
tsleep = 0.3


argv = sys.argv
n = len(argv)
if n >= 2:
    method = argv[1]
if n >= 3:
    x0s[0] = float(argv[2])
if n >= 4:
    h = float(argv[3])
if n >= 5:
    maxiter = int(argv[4])
if n >= 6:
    tol = float(argv[5])


def minimize_func(x):
    return (x[0] - 20.0 * x[0]**2 - 3.5 * x[0]**3 + 2.0 * x[0]**4) * sin(x[0])

# １次微分を定義するとcgやbfgsなどの勾配法を使える
def diff1(x):
    diff = x.copy()
    nvar = len(x)

    for i in range(nvar):
        xx = x.copy()
        xx[i] = x[i] - h
        ym = minimize_func(xx)
        
        xx[i] = x[i] + h
        yp = minimize_func(xx)
        diff[i] = (yp - ym) / 2.0 / h

    return diff


# callbackを使うと、最適化過程をモニターできる
# 引数は変数のリストだけが渡される
# 反復回数などはglobal変数で保持する
iter = 0
ax   = None
def callback(xk):
    global iter

    fmin = minimize_func(xk)
    print(f"callback {iter}: xk={xk}  func={fmin}")
    iter += 1

# グラフに更新した点を追加
    ax.plot(xk[0], minimize_func(xk), linestyle = '', marker = 'o', markersize = 5.0, markerfacecolor = 'red')

    plt.pause(tsleep)


#==========================================
# Main routine
#==========================================
def main():
    global x0s, tol, maxiter
    global ax
    
    print("")
    print(f"method        : {method}")
    print(f"initial values: {x0s}")
    print(f"tol           : {tol}")
    print(f"maxiter       : {maxiter}")

# グラフに表示する関数
    xgstep = (xgmax - xgmin) / (ngdata - 1)
    xg = [xgmin + xgstep * i for i in range(ngdata)]
    yini = [minimize_func([x]) for x in xg]

#関数のグラフ
    if fplot == 1:
        figure = plt.figure(figsize = (5, 5))
        ax = figure.add_subplot(1, 1, 1)
        ax.plot(xg, yini, color = 'black', linestyle = '-', linewidth = 0.5)
        ax.plot(x0s[0], minimize_func(x0s), linestyle = '', marker = 'o', markersize = 5.0, markerfacecolor = 'black')

        plt.pause(0.001)


    print("")
    print("Minimize:")
# 勾配法を使う場合はjacに一次微分のリストを返す関数を指定
# callbackを使うと、各iterationごとにcallbackが呼び出される    
    res = minimize(minimize_func, x0s, method = method, jac = diff1, callback = callback,
                tol = tol, options = {'maxiter':maxiter, "disp":True})
#    print("")
#    print(res)
    print("")
    if res.success:
        print(f"Function takes the minimum")
        print(f"   at y={res.fun}")
        print(f"   with x={res.x}")
        print(f" iteration: {res.nit}")
    else:
        print(f"Function did not converge")
        print(res)


    if fplot == 1:
        print("Press ENTER to terminate:", end = '')
        ret = input()


if __name__ == "__main__":
    main()
